'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.default = transformVisitors;

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _typeAnnotationIterator = require('./typeAnnotationIterator');

var _typeAnnotationIterator2 = _interopRequireDefault(_typeAnnotationIterator);

var _convert = require('./convert');

var _convert2 = _interopRequireDefault(_convert);

var _attachImport = require('./attachImport');

var _attachImport2 = _interopRequireDefault(_attachImport);

var _getTypeParameters = require('./getTypeParameters');

var _getTypeParameters2 = _interopRequireDefault(_getTypeParameters);

var _assert = require('assert');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function transformVisitors(context) {
  var shouldAssert = context.shouldAssert;
  return {
    Program: function Program(path) {
      if (context.shouldImport) {
        (0, _attachImport2.default)(context, path);
      }
    },
    'DeclareVariable|DeclareTypeAlias|DeclareFunction|DeclareClass|DeclareModule|InterfaceDeclaration': function DeclareVariableDeclareTypeAliasDeclareFunctionDeclareClassDeclareModuleInterfaceDeclaration(path) {
      var replacement = (0, _convert2.default)(context, path);
      path.replaceWith(replacement);
    },

    ImportDeclaration: {
      exit: function exit(path) {
        if (path.node.importKind !== 'type') {
          return;
        }
        path.node.importKind = 'value';
      }
    },
    ExportDeclaration: {
      exit: function exit(path) {
        if (path.node.exportKind !== 'type') {
          return;
        }
        path.node.exportKind = 'value';
      }
    },
    TypeAlias: function TypeAlias(path) {
      var replacement = (0, _convert2.default)(context, path);
      path.replaceWith(replacement);
    },
    TypeCastExpression: function TypeCastExpression(path) {
      var expression = path.get('expression');
      var typeAnnotation = path.get('typeAnnotation');
      if (shouldAssert && !expression.isIdentifier()) {
        path.replaceWith(t.callExpression(t.memberExpression((0, _convert2.default)(context, typeAnnotation), t.identifier('assert')), [expression.node]));
        return;
      }
      var name = expression.node.name;
      var binding = path.scope.getBinding(name);
      if (binding) {
        if (binding.path.isCatchClause()) {
          // special case typecasts for error handlers.
          path.parentPath.replaceWith(t.ifStatement(t.unaryExpression('!', t.callExpression(t.memberExpression((0, _convert2.default)(context, typeAnnotation), t.identifier('accepts')), [expression.node])), t.blockStatement([t.throwStatement(expression.node)])));
          return;
        } else if (name === 'reify') {
          if (typeAnnotation.isTypeAnnotation()) {
            var annotation = typeAnnotation.get('typeAnnotation');
            var isTypeWrapper = annotation.isGenericTypeAnnotation() && annotation.node.id.name === 'Type' && annotation.node.typeParameters && annotation.node.typeParameters.params && annotation.node.typeParameters.params.length === 1;
            if (isTypeWrapper) {
              path.replaceWith((0, _convert2.default)(context, annotation.get('typeParameters.params')[0]));
              return;
            }
          }
          path.replaceWith((0, _convert2.default)(context, typeAnnotation));
          return;
        }
      }

      if (!path.parentPath.isExpressionStatement()) {
        if (!shouldAssert) {
          return;
        }
        // this typecast is part of a larger expression, just replace the value inline.
        path.replaceWith(t.callExpression(t.memberExpression((0, _convert2.default)(context, typeAnnotation), t.identifier('assert')), [expression.node]));
        return;
      }

      var valueUid = path.scope.getData('valueUid:' + name);
      if (!valueUid) {
        valueUid = path.scope.generateUidIdentifier(name + 'Type');
        path.scope.setData('valueUid:' + name, valueUid);
        path.getStatementParent().insertBefore(t.variableDeclaration('let', [t.variableDeclarator(valueUid, (0, _convert2.default)(context, typeAnnotation))]));
      } else {
        path.getStatementParent().insertBefore(t.expressionStatement(t.assignmentExpression('=', valueUid, (0, _convert2.default)(context, typeAnnotation))));
      }
      if (shouldAssert) {
        path.replaceWith(t.callExpression(t.memberExpression(valueUid, t.identifier('assert')), [expression.node]));
      } else {
        path.replaceWith(expression.node);
      }
    },
    VariableDeclarator: function VariableDeclarator(path) {
      var id = path.get('id');
      if (!id.has('typeAnnotation')) {
        return;
      }
      if (!id.isIdentifier()) {
        (0, _assert.ok)(id.isArrayPattern() || id.isObjectPattern());
        var init = path.get('init');
        var wrapped = init.node;
        if (shouldAssert) {
          wrapped = t.callExpression(t.memberExpression((0, _convert2.default)(context, id.get('typeAnnotation')), t.identifier('assert')), [wrapped]);
        }
        if (wrapped !== init.node) {
          init.replaceWith(wrapped);
        }
        return;
      }
      var name = id.node.name;


      if (!path.has('init') || path.parentPath.node.kind !== 'const') {
        var valueUid = path.scope.generateUidIdentifier(name + 'Type');
        path.scope.setData('valueUid:' + name, valueUid);
        path.insertBefore(t.variableDeclarator(valueUid, (0, _convert2.default)(context, id.get('typeAnnotation'))));
        if (shouldAssert && path.has('init')) {
          var _wrapped = t.callExpression(t.memberExpression(valueUid, t.identifier('assert')), [path.get('init').node]);
          path.replaceWith(t.variableDeclarator(t.identifier(name), _wrapped));
        } else {
          id.replaceWith(t.identifier(name));
        }
      } else if (shouldAssert) {
        var _wrapped2 = t.callExpression(t.memberExpression((0, _convert2.default)(context, id.get('typeAnnotation')), t.identifier('assert')), [path.get('init').node]);
        path.replaceWith(t.variableDeclarator(t.identifier(name), _wrapped2));
      } else {
        id.replaceWith(t.identifier(name));
      }
    },
    AssignmentExpression: function AssignmentExpression(path) {
      var left = path.get('left');
      if (!shouldAssert || !left.isIdentifier()) {
        return;
      }
      var name = left.node.name;
      var valueUid = path.scope.getData('valueUid:' + name);
      if (!valueUid) {
        return;
      }
      var right = path.get('right');
      right.replaceWith(t.callExpression(t.memberExpression(valueUid, t.identifier('assert')), [right.node]));
    },
    Function: function Function(path) {
      var body = path.get('body');

      var definitions = [];
      var invocations = [];
      var typeParameters = (0, _getTypeParameters2.default)(path);
      var params = path.get('params');

      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = typeParameters[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var typeParameter = _step.value;
          var name = typeParameter.node.name;

          var args = [t.stringLiteral(name)];
          if (typeParameter.has('bound')) {
            args.push((0, _convert2.default)(context, typeParameter.get('bound')));
          }
          definitions.push(t.variableDeclaration('const', [t.variableDeclarator(t.identifier(name), context.call.apply(context, ['typeParameter'].concat(args)))]));
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }

      var shouldShadow = false;

      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = params[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var param = _step2.value;

          var argumentIndex = +param.key;
          var assignmentRight = void 0;
          if (param.isAssignmentPattern()) {
            assignmentRight = param.get('right');
            param = param.get('left');
          }
          if (!param.has('typeAnnotation')) {
            continue;
          }
          var typeAnnotation = param.get('typeAnnotation');

          if (param.isObjectPattern() || param.isArrayPattern()) {
            if (shouldAssert) {
              shouldShadow = true;

              var _args = [t.stringLiteral('arguments[' + argumentIndex + ']'), (0, _convert2.default)(context, typeAnnotation)];
              if (param.has('optional')) {
                _args.push(t.booleanLiteral(true));
              }

              var ref = t.memberExpression(t.identifier('arguments'), t.numericLiteral(argumentIndex), true);

              var expression = t.expressionStatement(t.callExpression(t.memberExpression(context.call.apply(context, ['param'].concat(_args)), t.identifier('assert')), [ref]));
              if (assignmentRight) {
                invocations.push(t.ifStatement(t.binaryExpression('!==', ref, t.identifier('undefined')), t.blockStatement([expression])));
              } else {
                invocations.push(expression);
              }
            }
          } else {
            var _name = param.node.name;
            var methodName = 'param';
            if (param.isRestElement()) {
              methodName = 'rest';
              _name = param.node.argument.name;
            } else if (!param.isIdentifier()) {
              continue;
            }
            var valueUid = body.scope.generateUidIdentifier(_name + 'Type');
            body.scope.setData('valueUid:' + _name, valueUid);
            definitions.push(t.variableDeclaration('let', [t.variableDeclarator(valueUid, (0, _convert2.default)(context, typeAnnotation))]));
            if (shouldAssert) {
              var _args2 = [t.stringLiteral(_name), valueUid];
              if (param.has('optional')) {
                _args2.push(t.booleanLiteral(true));
              }
              invocations.push(t.expressionStatement(t.callExpression(t.memberExpression(context.call.apply(context, [methodName].concat(_args2)), t.identifier('assert')), [t.identifier(_name)])));
            }
          }
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }

      if (path.has('returnType')) {
        var returnType = path.get('returnType');
        if (returnType.type === 'TypeAnnotation') {
          returnType = returnType.get('typeAnnotation');
        }
        var returnTypeParameters = (0, _getTypeParameters2.default)(returnType);
        if (returnType.isGenericTypeAnnotation() && returnTypeParameters.length > 0) {
          // If we're in an async function, make the return type the promise resolution type.
          if (path.node.async) {
            // @todo warn if identifier is not Promise ?
            returnType = (0, _getTypeParameters2.default)(returnType)[0];
          } else if (path.node.generator) {
            var yieldType = returnTypeParameters[0];
            returnType = returnTypeParameters[1];
            var nextType = returnTypeParameters[2];
            var yieldTypeUid = body.scope.generateUidIdentifier('yieldType');
            body.scope.setData('yieldTypeUid', yieldTypeUid);
            definitions.push(t.variableDeclaration('const', [t.variableDeclarator(yieldTypeUid, (0, _convert2.default)(context, yieldType))]));
            var nextTypeUid = body.scope.generateUidIdentifier('nextType');
            body.scope.setData('nextTypeUid', nextTypeUid);
            definitions.push(t.variableDeclaration('const', [t.variableDeclarator(nextTypeUid, (0, _convert2.default)(context, nextType))]));
          }
        }
        var returnTypeUid = body.scope.generateUidIdentifier('returnType');
        body.scope.setData('returnTypeUid', returnTypeUid);
        definitions.push(t.variableDeclaration('const', [t.variableDeclarator(returnTypeUid, context.call('return', (0, _convert2.default)(context, returnType)))]));
      }
      if (definitions.length > 0 || invocations.length > 0) {

        if (shouldShadow && path.isArrowFunctionExpression()) {
          path.arrowFunctionToShadowed();
          path.get('body').unshiftContainer('body', definitions.concat(invocations));
        } else {
          body.unshiftContainer('body', definitions.concat(invocations));
        }
      }
    },
    ReturnStatement: function ReturnStatement(path) {
      var fn = path.scope.getFunctionParent().path;
      if (!shouldAssert || !fn.has('returnType')) {
        return;
      }
      var returnTypeUid = path.scope.getData('returnTypeUid');

      var argument = path.get('argument');
      argument.replaceWith(t.callExpression(t.memberExpression(returnTypeUid, t.identifier('assert')), argument.node ? [argument.node] : []));
    },
    YieldExpression: function YieldExpression(path) {
      var fn = path.scope.getFunctionParent().path;
      if (!shouldAssert || !fn.has('returnType')) {
        return;
      }
      if (context.visited.has(path.node)) {
        return;
      }
      var yieldTypeUid = path.scope.getData('yieldTypeUid');
      var nextTypeUid = path.scope.getData('nextTypeUid');

      var argument = path.get('argument');
      var replacement = void 0;
      if (path.node.delegate) {
        replacement = t.yieldExpression(t.callExpression(context.call('wrapIterator', yieldTypeUid), argument.node ? [argument.node] : []), true);
      } else {
        replacement = t.yieldExpression(t.callExpression(t.memberExpression(yieldTypeUid, t.identifier('assert')), argument.node ? [argument.node] : []));
      }

      context.visited.add(replacement);
      if (path.parentPath.isExpressionStatement()) {
        path.replaceWith(replacement);
      } else {
        path.replaceWith(t.callExpression(t.memberExpression(nextTypeUid, t.identifier('assert')), [replacement]));
      }
    },


    Class: {
      enter: function enter(path) {
        var superTypeParameters = path.has('superTypeParameters') ? path.get('superTypeParameters.params') : [];
        var hasSuperTypeParameters = superTypeParameters.length > 0;
        if (path.has('superClass') && isReactComponentClass(path.get('superClass'))) {
          var annotation = hasSuperTypeParameters ? superTypeParameters[1] : getClassPropertyAnnotation(path, 'props');

          if (annotation) {
            var body = path.get('body');
            var propTypes = t.classProperty(t.identifier('propTypes'), context.call('propTypes', (0, _convert2.default)(context, annotation)));
            propTypes.static = true;
            body.unshiftContainer('body', propTypes);
          }
        }
      },
      exit: function exit(path) {
        var typeParameters = (0, _getTypeParameters2.default)(path);
        var superTypeParameters = path.has('superTypeParameters') ? path.get('superTypeParameters.params') : [];
        var hasTypeParameters = typeParameters.length > 0;
        var hasSuperTypeParameters = superTypeParameters.length > 0;
        if (!hasTypeParameters && !hasSuperTypeParameters) {
          // Nothing to do here.
          return;
        }

        var _path$get$filter = path.get('body.body').filter(function (item) {
          return item.node.kind === 'constructor';
        }),
            _path$get$filter2 = _slicedToArray(_path$get$filter, 1),
            constructor = _path$get$filter2[0];

        var typeParametersUid = t.identifier(path.scope.getData('typeParametersUid'));
        if (path.has('superClass')) {
          var body = constructor.get('body');

          var trailer = [];
          if (hasTypeParameters) {
            body.unshiftContainer('body', t.variableDeclaration('const', [t.variableDeclarator(typeParametersUid, t.objectExpression(typeParameters.map(function (typeParameter) {
              return t.objectProperty(t.identifier(typeParameter.node.name), (0, _convert2.default)(context, typeParameter));
            })))]));

            var thisTypeParameters = t.memberExpression(t.thisExpression(), context.symbol('TypeParameters'), true);

            trailer.push(t.ifStatement(thisTypeParameters, t.blockStatement([t.expressionStatement(t.callExpression(t.memberExpression(t.identifier('Object'), t.identifier('assign')), [thisTypeParameters, typeParametersUid]))]), t.blockStatement([t.expressionStatement(t.assignmentExpression('=', thisTypeParameters, typeParametersUid))])));
          }

          if (hasSuperTypeParameters) {
            trailer.push(t.expressionStatement(context.call.apply(context, ['bindTypeParameters', t.thisExpression()].concat(_toConsumableArray(superTypeParameters.map(function (item) {
              return (0, _convert2.default)(context, item);
            }))))));
          }
          getSuperStatement(body).insertAfter(trailer);
        } else {
          constructor.get('body').unshiftContainer('body', t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.thisExpression(), context.symbol('TypeParameters'), true), t.objectExpression(typeParameters.map(function (typeParameter) {
            return t.objectProperty(t.identifier(typeParameter.node.name), (0, _convert2.default)(context, typeParameter));
          })))));
        }
        var staticMethods = path.get('body.body').filter(function (item) {
          return item.isClassMethod() && item.node.static;
        });

        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;

        try {
          for (var _iterator3 = staticMethods[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
            var method = _step3.value;

            method.get('body').unshiftContainer('body', t.variableDeclaration('const', [t.variableDeclarator(typeParametersUid, t.objectExpression(typeParameters.map(function (typeParameter) {
              return t.objectProperty(t.identifier(typeParameter.node.name), (0, _convert2.default)(context, typeParameter));
            })))]));
          }
        } catch (err) {
          _didIteratorError3 = true;
          _iteratorError3 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion3 && _iterator3.return) {
              _iterator3.return();
            }
          } finally {
            if (_didIteratorError3) {
              throw _iteratorError3;
            }
          }
        }
      }
    },

    ClassProperty: function ClassProperty(path) {
      if (!path.has('typeAnnotation')) {
        return;
      }
      var typeAnnotation = path.get('typeAnnotation');
      var decorator = void 0;
      if (annotationReferencesClassEntity(context, typeAnnotation)) {
        decorator = t.decorator(context.call('decorate', t.functionExpression(null, [], t.blockStatement([t.returnStatement((0, _convert2.default)(context, typeAnnotation))]))));
      } else {
        decorator = t.decorator(context.call('decorate', (0, _convert2.default)(context, typeAnnotation)));
      }
      if (!path.has('decorators')) {
        path.node.decorators = [];
      }
      path.pushContainer('decorators', decorator);
    }
  };
}

function isReactComponentClass(path) {
  if (path.isIdentifier()) {
    return path.node.name === path.scope.getData('reactComponentClass') || path.node.name === path.scope.getData('reactPureComponentClass');
  } else if (path.isMemberExpression() && !path.node.computed) {
    var object = path.get('object');
    var property = path.get('property');
    if (!object.isIdentifier() || object.node.name !== path.scope.getData('reactLib')) {
      return false;
    }
    return property.isIdentifier() && (property.node.name === 'Component' || property.node.name === 'PureComponent');
  } else {
    return false;
  }
}

function getClassPropertyAnnotation(path, name) {
  var _iteratorNormalCompletion4 = true;
  var _didIteratorError4 = false;
  var _iteratorError4 = undefined;

  try {
    for (var _iterator4 = path.get('body.body')[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
      var item = _step4.value;

      if (item.isClassProperty() && item.node.key.name === name) {
        return item.get('typeAnnotation');
      }
    }
  } catch (err) {
    _didIteratorError4 = true;
    _iteratorError4 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion4 && _iterator4.return) {
        _iterator4.return();
      }
    } finally {
      if (_didIteratorError4) {
        throw _iteratorError4;
      }
    }
  }
}

function annotationReferencesClassEntity(context, annotation) {
  var _iteratorNormalCompletion5 = true;
  var _didIteratorError5 = false;
  var _iteratorError5 = undefined;

  try {
    for (var _iterator5 = (0, _typeAnnotationIterator2.default)(annotation)[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
      var item = _step5.value;

      if (item.type !== 'Identifier') {
        continue;
      }
      var entity = context.getEntity(item.node.name, annotation);
      if (entity && entity.isClassTypeParameter) {
        return true;
      } else if (entity && entity.isValue && !entity.isGlobal) {
        return true;
      }
    }
  } catch (err) {
    _didIteratorError5 = true;
    _iteratorError5 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion5 && _iterator5.return) {
        _iterator5.return();
      }
    } finally {
      if (_didIteratorError5) {
        throw _iteratorError5;
      }
    }
  }

  return false;
}

function getSuperStatement(block) {
  var found = void 0;
  block.traverse({
    Super: function Super(path) {
      found = path.getStatementParent();
    }
  });
  (0, _assert.ok)(found, "Constructor of sub class must contain super().");
  return found;
}